import glob
import os
import re
import sys
import warnings
from importlib import import_module

from .conf import conf


class BaseFinder(object):
    def find(self, name):
        raise NotImplementedError()


class FilesystemFinder(BaseFinder):
    @property
    def directories(self):
        return conf.JSTEMPLATE_DIRS

    @property
    def extensions(self):
        return conf.JSTEMPLATE_EXTS

    def find(self, name):
        matches = {}

        for directory in self.directories:
            for extension in self.extensions:
                self._update_matches(matches, name, directory, extension)

        return matches.items()

    def _update_matches(self, matches, pathspec, directory, extension):
        """
        Add names and paths corresponding to the given glob pattern (pathspec)
        to the mapping (matches).  Do not override existing matching entries.
        """

        lenext = len(extension) + 1  # The +1 is for the '.'
        lendir = len(directory)

        # Catch the ending '/', if it's not already accounted for
        if not directory.endswith(os.path.sep):
            lendir += 1

        normdir = os.path.normpath(directory)

        pathname = os.path.join(directory, pathspec + "." + extension)
        pathmatches = glob.glob(pathname)

        for filepath in pathmatches:
            # Remove the template directory name and the extension to
            # get the template name
            matchname = filepath[lendir:-lenext]

            # Add matchname => filepath to the mapping.  Do not override.
            if matchname not in matches:
                absfilepath = os.path.abspath(filepath)
                if absfilepath.startswith(normdir):
                    matches[matchname] = absfilepath


class FilesystemRegexFinder(BaseFinder):
    @property
    def directories(self):
        return conf.JSTEMPLATE_DIRS

    @property
    def extensions(self):
        return conf.JSTEMPLATE_EXTS

    def find(self, name):
        matches = {}

        try:
            pattern = re.compile(name + ".(?:" + '|'.join(self.extensions) + ")$")
        except re.error:
            return []

        # Bail if there are no capturing groups in the pattern
        if pattern.groups == 0:
            return []

        for directory in self.directories:
            self._update_matches(matches, directory, pattern)

        return matches.items()

    def _update_matches(self, matches, directory, pattern):
        # Walk recursively under directory and search for matches
        for root, dirs, files in os.walk(directory):
            # Build relative path to current directory, omit .
            relative_root = os.path.relpath(root, directory).replace('.', '', 1)
            for filename in files:
                # Build relative path to file
                relative_filepath = os.path.join(relative_root, filename)
                match = pattern.match(relative_filepath)
                if match is not None and match.group(1) not in matches:
                    matches[match.group(1)] = os.path.join(directory, relative_filepath)


def _get_app_template_dirs():
    fs_encoding = sys.getfilesystemencoding() or sys.getdefaultencoding()
    ret = []
    for app in conf.INSTALLED_APPS:
        try:
            mod = import_module(app)
        except ImportError:
            warnings.warn('Installed app %s is not an importable Python module and will be ignored' % app)
            continue
        app_dir = os.path.dirname(os.path.abspath(mod.__file__))
        for dirname in conf.JSTEMPLATE_APP_DIRNAMES:
            template_dir = os.path.join(app_dir, dirname)
            if os.path.isdir(template_dir):
                ret.append(template_dir)
    return ret



# At import time, cache the app directories to search.
app_template_dirs = _get_app_template_dirs()



class AppFinder(FilesystemFinder):
    @property
    def directories(self):
        return app_template_dirs



class AppRegexFinder(FilesystemRegexFinder):
    @property
    def directories(self):
        return app_template_dirs



from django.core.exceptions import ImproperlyConfigured
from importlib import import_module

from .conf import conf


def find(name):
    all_matches = {}

    for finder in finders:
        matches = finder.find(name)

        # <finder>.find may return a single string.  The name of the template
        # will then be the name given to 'find'
        if isinstance(matches, str):
            filepath = matches
            if name not in all_matches:
                all_matches[name] = filepath

        # None valus should be ignored
        elif matches is None:
            continue

        # Otherwise, matches should be a list of (name, filepath) pairs
        else:
            try:
                for matchname, filepath in matches:
                    if matchname not in all_matches:
                        all_matches[matchname] = filepath
            except ValueError:
                import pdb; pdb.set_trace()

    if len(all_matches) == 0:
        raise JSTemplateNotFound(name)

    return all_matches.items()


def preprocess(content):
    for preprocessor in preprocessors:
        content = preprocessor.process(content)

    return content


def _get_classes(dotted_paths):
    ret = []
    for path in dotted_paths:
        modpath, cls_name = path.rsplit(".", 1)
        try:
            mod = import_module(modpath)
        except ImportError as e:
            raise ImproperlyConfigured(
                "ImportError %s: %s" % (modpath, e.args[0]))

        try:
            cls = getattr(mod, cls_name)
        except AttributeError as e:
            raise ImproperlyConfigured(
                "AttributeError %s: %s" % (cls_name, e.args[0]))

        ret.append(cls())

    return ret


def _get_finders():
    return _get_classes(conf.JSTEMPLATE_FINDERS)


def _get_preprocessors():
    return _get_classes(conf.JSTEMPLATE_PREPROCESSORS)


# Instantiate finders
finders = _get_finders()
preprocessors = _get_preprocessors()


class JSTemplateNotFound(Exception):
    pass



from django import template
from ..conf import conf
from ..loading import find, preprocess, JSTemplateNotFound


class BaseJSTemplateNode(template.Node):
    preprocessors = conf.JSTEMPLATE_PREPROCESSORS

    def __init__(self, name, *args):
        self.name = template.Variable(name)
        self.args = args

    def find_template_matches(self, name):
        return find(name)

    def preprocess(self, content):
        return preprocess(content)

    def render(self, context):
        name = self.name.resolve(context)

        try:
            matches = self.find_template_matches(name)
        except JSTemplateNotFound:
            if conf.DEBUG: raise
            else: return ''
        else:
            return ''.join([self.render_file(context, matchname, filepath)
                            for (matchname, filepath) in matches])

    def render_file(self, context, resolved_name, filepath):
        try:
            content = self.read_template_file_contents(filepath)
            content = self.preprocess(content)
            output = self.generate_node_text(resolved_name, content)

        except IOError:
            output = ""
            if conf.DEBUG:
                raise

        return output

    def read_template_file_contents(self, filepath):
        # As of Django 2.2, the FILE_CHARSET setting is deprecated, and as of
        # Django 3.1, UTF-8 is always expected.
        encoding = getattr(conf, 'FILE_CHARSET', 'utf-8')

        with open(filepath, "r", encoding=encoding) as fp:
            template_text = fp.read()
        template_text = self.preprocess(template_text)
        return template_text

    def generate_node_text(self, resolved_name, file_content):
        raise NotImplementedError()


def jstemplate_tag_helper(tagname, TagNodeClass, parser, token):
    """
    Finds the JavaScript template or templates specified in the tag arguments
    and uses the appropriate tag node to render it.

    """
    bits = token.contents.split()
    if len(bits) < 2:
        raise template.TemplateSyntaxError(
            "'%s' tag takes at least one argument: the name/id of "
            "the template, or a pattern matching a set of templates. "
            % tagname)
    return TagNodeClass(*bits[1:])



import os.path

from django.test import override_settings, TestCase

from jstemplate.management.commands import makemessages

DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "project", "project", "jstemplates")

class MonkeyPatchedTemplatizeTest (TestCase):

    @override_settings(JSTEMPLATE_DIRS=[DIR])
    def test_translateable_strings_are_discovered(self):
        filepath = os.path.join(DIR, 'my-template.html')
        result = makemessages.templatize(None, filepath)
        self.assertEqual(result, '''_("""Here is
      a multi-line
      string
""")
_("""it's Mustache's turn to do i18n""")
_("""this is a \\"<b>string</b>\\" with <i>tags</i>""")
_("""this is a {{ string }} {{# with }} variables {{/ with }}""")''')



from django.conf import settings


class Configuration(object):
    def __init__(self, **kwargs):
        self.defaults = kwargs

    def __getattr__(self, k):
        try:
            return getattr(settings, k)
        except AttributeError:
            if k in self.defaults:
                return self.defaults[k]
            raise


conf = Configuration(
    JSTEMPLATE_FINDERS=[
        "jstemplate.finders.FilesystemFinder",
        "jstemplate.finders.AppFinder",
        "jstemplate.finders.FilesystemRegexFinder",
        "jstemplate.finders.AppRegexFinder",
        ],
    JSTEMPLATE_DIRS=[],
    JSTEMPLATE_EXTS=["mustache", "html"],
    JSTEMPLATE_APP_DIRNAMES=["jstemplates"],
    JSTEMPLATE_I18N_TAGS=["_", "i18n"],
    JSTEMPLATE_PREPROCESSORS=['jstemplate.preprocessors.I18nPreprocessor'],
    )



import os

from django.test import override_settings, TestCase

from unittest.mock import patch



__all__ = ["BaseFinderTest", "FilesystemFinderTest", "AppFinderTest",
           "FilesystemRegexFinderTest"]


here = os.path.abspath(os.path.dirname(os.path.abspath(__file__)))



class BaseFinderTest(TestCase):
    @property
    def finder(self):
        from jstemplate.finders import BaseFinder
        return BaseFinder()


    def test_find_not_implemented(self):
        with self.assertRaises(NotImplementedError):
            self.finder.find("name")



class FilesystemFinderTest(TestCase):
    @property
    def finder(self):
        from jstemplate.finders import FilesystemFinder
        return FilesystemFinder()


    @override_settings(JSTEMPLATE_DIRS=["/one/path", "/another/path"])
    def test_directories(self):
        self.assertEqual(
            self.finder.directories,
            ["/one/path", "/another/path"])


    @override_settings(JSTEMPLATE_EXTS=["thtml"])
    def test_extensions(self):
        self.assertEqual(
            self.finder.extensions,
            ["thtml"])


    @override_settings(JSTEMPLATE_DIRS=[os.path.join(here, "templates")])
    def test_find_html_file(self):
        self.assertEqual(
            list(self.finder.find("testtemplate")),
            [("testtemplate", os.path.join(here, "templates", "testtemplate.html"))])


    @override_settings(JSTEMPLATE_DIRS=[os.path.join(here, "templates")])
    def test_find_mustache_file(self):
        # This will not only demonstrate that it finds .mustache files, but also
        # that .mustache files take precedence over .html files.
        self.assertEqual(
            list(self.finder.find("othertesttemplate")),
            [("othertesttemplate", os.path.join(here, "templates", "othertesttemplate.mustache"))])


    @override_settings(JSTEMPLATE_DIRS=[os.path.join(here, "templates")])
    def test_find_many_files(self):
        self.assertEqual(
            set(self.finder.find("many*")),
            set([("manytemplates1", os.path.join(here, "templates", "manytemplates1.html")),
                 ("manytemplates2", os.path.join(here, "templates", "manytemplates2.mustache")),
                 ("manytemplates3", os.path.join(here, "templates", "manytemplates3.html"))])
        )


    @override_settings(JSTEMPLATE_DIRS=[os.path.join(here, "templates")])
    def test_mustache_file_takes_precedence_over_html(self):
        self.assertEqual(
            list(self.finder.find("bothtesttemplate")),
            [("bothtesttemplate", os.path.join(here, "templates", "bothtesttemplate.mustache"))])


    @override_settings(
        JSTEMPLATE_DIRS=[os.path.join(here, "..", "tests", "templates")])
    def test_find_non_normalized_dir(self):
        self.assertEqual(
            list(self.finder.find("testtemplate")),
            [("testtemplate", os.path.join(here, "templates", "testtemplate.html"))])


    @override_settings(JSTEMPLATE_DIRS=[os.path.join(here, "templates")])
    def test_find_non_existing(self):
        self.assertEqual(list(self.finder.find("doesntexist")), [])


class FilesystemRegexFinderTest(TestCase):
    @property
    def finder(self):
        from jstemplate.finders import FilesystemRegexFinder
        return FilesystemRegexFinder()


    @override_settings(JSTEMPLATE_DIRS=[os.path.join(here, "templates")])
    def test_find_file(self):
        self.assertEqual(
            list(self.finder.find("(test.*)")),
            [("testtemplate", os.path.join(here, "templates", "testtemplate.html"))])


    @override_settings(JSTEMPLATE_DIRS=[os.path.join(here, "templates")])
    def test_find_file(self):
        self.assertEqual(
            list(self.finder.find("(test.*)")),
            [("testtemplate", os.path.join(here, "templates", "testtemplate.html"))])


    @override_settings(JSTEMPLATE_DIRS=[os.path.join(here, "templates")])
    def test_requires_match_group(self):
        self.assertEqual(
            list(self.finder.find("test.*")), [])


    @override_settings(JSTEMPLATE_DIRS=[os.path.join(here, "templates")])
    def test_requires_exact_match_for_extensions(self):
        self.assertEqual(
            list(self.finder.find("(wrong_extension)")), [])


    @override_settings(JSTEMPLATE_DIRS=[os.path.join(here, "templates")])
    def test_find_many_files(self):
        self.assertEqual(
            set(self.finder.find("(many.*)")),
            set([("manytemplates1", os.path.join(here, "templates", "manytemplates1.html")),
                 ("manytemplates2", os.path.join(here, "templates", "manytemplates2.mustache")),
                 ("manytemplates3", os.path.join(here, "templates", "manytemplates3.html"))])
        )


    @override_settings(JSTEMPLATE_DIRS=[os.path.join(here, "templates")])
    def test_find_files_in_subdirectories(self):
        self.assertEqual(
            set(self.finder.find('(subdirectory.*)')),
            set([('subdirectory/subdirectory2/testtemplate', os.path.join(here, 'templates', 'subdirectory', 'subdirectory2', 'testtemplate.html')),
                 ('subdirectory/testtemplate', os.path.join(here, 'templates', 'subdirectory', 'testtemplate.html'))]))


    @override_settings(JSTEMPLATE_DIRS=[os.path.join(here, "templates")])
    def test_fails_silently_on_invalid_reges(self):
        self.assertEqual(self.finder.find("*"), [])

class AppFinderTest(TestCase):
    @property
    def finder(self):
        from jstemplate.finders import AppFinder
        return AppFinder()


    def test_directories(self):
        with patch(
            "jstemplate.finders.app_template_dirs",
            [os.path.join(here, "templates")]):
            dirs = self.finder.directories

        self.assertEqual(dirs, [os.path.join(here, "templates")])


    @property
    def func(self):
        from jstemplate.finders import _get_app_template_dirs
        return _get_app_template_dirs


    @override_settings(
        INSTALLED_APPS=["jstemplate.tests"],
        JSTEMPLATE_APP_DIRNAMES=["templates", "jstemplates"])
    def test_get_app_template_dirs(self):
        self.assertEqual(self.func(), [os.path.join(here, "templates")])


    def test_bad_app(self):
        with self.assertRaises(ModuleNotFoundError):
            # Adding the non-existent app to INSTALLED_APPS triggers a
            # ModuleNotFoundError
            with override_settings(INSTALLED_APPS=["jstemplate.nonexistent"]):
                pass



from django import template
from .base import BaseJSTemplateNode, jstemplate_tag_helper
from ..conf import conf

register = template.Library()


class MustacheJSNode(BaseJSTemplateNode):
    def generate_node_text(self, resolved_name, file_content):
        output = file_content
        output = output.replace('\\', r'\\')
        output = output.replace('\n', r'\n')
        output = output.replace("'", r"\'")

        output = ("<script>Mustache.TEMPLATES=Mustache.TEMPLATES||{};"
                    + "Mustache.TEMPLATES['{0}']='".format(resolved_name)
                    + output + "';</script>")

        return output



@register.tag
def mustachejs(parser, token):
    """
    Finds the MustacheJS template for the given name and renders it surrounded by
    the requisite MustacheJS <script> tags.

    """
    return jstemplate_tag_helper('mustachejs', MustacheJSNode,
                                 parser, token)



from django import template
from .base import BaseJSTemplateNode, jstemplate_tag_helper



register = template.Library()



class RawJSTemplateNode(BaseJSTemplateNode):
    def generate_node_text(self, resolved_name, file_content):
        return file_content



@register.tag
def rawjstemplate(parser, token):
    """
    Prints out raw mustache content

    """
    return jstemplate_tag_helper('rawjstemplate', RawJSTemplateNode,
                                 parser, token)



from django import template
from .base import BaseJSTemplateNode, jstemplate_tag_helper



register = template.Library()



class MustacheICanHazNode(BaseJSTemplateNode):
    def generate_node_text(self, resolved_name, file_content):
        output = ('<script type="text/html" id="{0}">'.format(resolved_name)
                    + file_content + '</script>')
        return output



@register.tag
def icanhazjs(parser, token):
    """
    Finds the Mustache template for the given name and renders it surrounded by
    the requisite ICanHaz <script> tags.

    """
    return jstemplate_tag_helper('icanhazjs', MustacheICanHazNode,
                                 parser, token)



"""In order to avoid having to send our project's translation mapping to the
client, we have built-in the ability to preprocess i18n tags in the mustache
templates.

There aren't any nice solutions here.  The code behind ``makemessages``
unfortunately isn't extensible, so we can:

  * Duplicate the command + code behind it.
  * Offer a separate command for Mustache tag extraction.
  * Try to get Django to offer hooks into makemessages().
  * Monkey-patch.

We are currently doing that last thing. In this case we override the templatize
method.  Templatize takes a template, extracts the translatable strings (along
with desired metadata), and generates a file that xgettext knows how to parse,
e.g. a file with Python syntax.  We override this function to find
Mustache-tagged strings if the file that we are templatizing is in one of the
paths found by the active MUSTACHEJS_FINDERS.
"""

import os
import re
import sys
from itertools import chain

from django.utils.translation \
    import templatize as base_templatize

from jstemplate.loading import find, preprocess
from jstemplate.preprocessors import I18nPreprocessor


def templatize(src, origin=None):
    """
    Override the templatize function to extract translatable strings from
    jstemplates.
    """
    # Get all the paths that we know about
    paths = [os.path.abspath(path) for name, path in find('(.*)')]

    # Hijack the process if the file we're talking about is in one of the
    # finder paths.
    if origin and os.path.abspath(origin) in paths:
        with open(origin) as template_file:

            # Load the template content.
            content = template_file.read()

            # Find all the translatable strings.
            processor = I18nPreprocessor()
            short_pattern = re.compile(processor.short_trans_re)
            long_pattern = re.compile(processor.long_trans_re, re.DOTALL)
            strings = set()

            for match in chain(re.finditer(short_pattern, content),
                               re.finditer(long_pattern, content)):
                strings.add(match.group('msg'))

            def escape(s):
                s = s.replace('\\', '\\\\')
                s = s.replace('"', r'\"')
                return s

            # Build a string that looks like a Python file that's ready to be
            # translated.
            translatable = '\n'.join(sorted('_("""{0}""")'.format(escape(string))
                                            for string in strings))

            return translatable

    # If the file isn't in one of our paths, then delegate to the original
    # method.
    else:
        return base_templatize(src, origin=origin)


# ============================================================================
#
# Monkey-patch django.utils.translation.templatize to use the mustachejs
# version when dealing with a file that is in a mustachejs finder directory.
# Patch it globally so that any other commands that also patch the function
# will inherit this functionality.

# Since Django 1.11, the django.core.management.commands.makemessages module
# has imported templatize from django.utils.translation directly.
# https://github.com/django/django/commit/5fb22b4d4c664e14b7734550afc258ae97a4f8be#diff-f6dcf84eeecbbfbad63c316715f09e3eaf55784c75295e97a7fa57f424a52205R24
import django.core.management.commands.makemessages as base_makemessages
base_makemessages.templatize = templatize

# ============================================================================


class Command (base_makemessages.Command):
    help = ('Adds the translatable strings from the js templates to the '
            'messages')

    def handle_noargs(self, *args, **options):
        return super(Command, self).handle_noargs(*args, **options)